@@ -8,7 +8,7 @@ use overload
     '""'     => sub { shift->{_value} },
     fallback => 1;
 
-our $VERSION = '1.0.11';
+our $VERSION = '1.1.0';
 
 
 require XS::Loader;
@@ -21,7 +21,10 @@ my $handler = sub {
     die($wrapped_ex) if ($wrapped_ex);
 };
 
+our $decorator = \&default_decorator;
+
 sub install {
+    $decorator = (shift) // \&default_decorator;
     $SIG{__DIE__} = $handler;
 };
 
@@ -31,5 +34,11 @@ sub new {
     return bless $obj => $class;
 }
 
+sub default_decorator {
+    my $args = shift;
+    my $r = join(', ', map { defined($_) ? "$_" : 'undef' } @$args);
+    return "($r)";
+}
+
 
 1;
